#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <cuda_runtime.h>
#include <cublas_v2.h>
#include <magma.h>

#define	ITER	6

#include "test-common.h"
#include "magma.h"


void
set_matrix( enum _FUNCS func,
            int N, TYPE *A, int LDA, TYPE *B, TYPE *C, TYPE *D )
{
  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      A[j+i*LDA]=CONST(0);
    }
  }
  for(int i=0;i<N*1;i++) B[i]=CONST(1);
  for(int i=0;i<N*1;i++) C[i]=CONST(0);
  for(int i=0;i<N*1;i++) D[i]=CONST(0);

#define	BB		(7)

  if (func==GEMV_N) {
    for(int kk=1;kk<=BB;kk++) {
      for(int k=0;k<N;k++) {
        int j=rand()%N; int i=k;
        A[i+j*LDA] += CONST(kk*kk*N);
      }
    }
  }
  if (func==GEMV_T) {
    for(int kk=1;kk<=BB;kk++) {
      for(int k=0;k<N;k++) {
        int i=rand()%N; int j=k;
        A[i+j*LDA] += CONST(kk*kk*N);
      }
    }
  }
#if TYPE_DOUBLE || TYPE_FLOAT
  if (func==SYMV_U||func==SYMV_L)
#else
    if (func==HEMV_U||func==HEMV_L)
#endif
      {
        int	*col = (int *)malloc(sizeof(int)*N);
        for(int kk=1;kk<=BB;kk++) {
          for(int k=0;k<N;k++) {
            col[k] = k;
          }
          for(int k=0;k<N;k++) {
            int i1=rand()%N;
            int i2=rand()%N;
            int j=col[i1]; col[i1]=col[i2]; col[i2]=j;
          }
          for(int k=0;k<N;k++) {
            int i=k, j=col[i];
#if TYPE_DOUBLE || TYPE_FLOAT
            A[i+j*LDA] += CONST(kk*kk*N);
            A[j+i*LDA] += CONST(kk*kk*N);
#elif TYPE_DOUBLE_COMPLEX
            A[i+j*LDA] +=     (make_cuDoubleComplex(kk*kk*N,1));
            A[j+i*LDA] += Conj(make_cuDoubleComplex(kk*kk*N,1));
#elif TYPE_FLOAT_COMPLEX
            A[i+j*LDA] +=     (make_cuFloatComplex(kk*kk*N,1));
            A[j+i*LDA] += Conj(make_cuFloatComplex(kk*kk*N,1));
#endif
          }
        }
        free(col);
      }

  for(int k=0;k<N/BB;k++) {
    while ( 1 ) {
      int j=rand()%N;
      if ( B[j] == CONST(1) ) {
        B[j] = CONST(0);
        break;
      }
    }
  }
  for(int k=0;k<N/BB;k++) {
    while ( 1 ) {
      int j=rand()%N;
      if ( B[j] == CONST(1) ) {
        B[j] = CONST(-1);
        break;
      }
    }
  }
}


int
main(int argc, char *argv[])
{
  TYPE	alpha = CONST(3.0);
  TYPE	beta  = CONST(0.0);//4.0;
  TYPE	*A, *B, *C, *D;
  TYPE	*dA, *dB, *dC, *dD;
  int	N, M, K;
  int	LDA, LDB, LDC;
  int	i;
  char	mode;
  int	blk;
  FILE	*fp;
  int	id = get_gpu_id();


  enum _FUNCS func = GEMV_N;
  char	*func_name = (char *)"";

  magma_init( );
  cudaSetDevice( id );

  cublasStatus_t stat;
  cublasHandle_t handle;
  cublasAtomicsMode_t atomic_mode = CUBLAS_ATOMICS_ALLOWED;
  stat = cublasCreate( &handle );
  cublasSetAtomicsMode( handle, atomic_mode );






  {

    char *argv0 = argv[0];
    if ( strrchr(argv0, '/') != argv0 ) {
      argv0 = strrchr(argv0,'/')+1;
    }
    if ( 0 == strcmp(argv0,TEST_GEMV_N) ) {
      func      = GEMV_N;
      func_name = (char *)NAME_GEMV_N;
    }
    if ( 0 == strcmp(argv0,TEST_GEMV_T) ) {
      func      = GEMV_T;
      func_name = (char *)NAME_GEMV_T;
    }
#if TYPE_DOUBLE || TYPE_FLOAT
    if ( 0 == strcmp(argv0,TEST_SYMV_U) ) {
      func      = SYMV_U;
      func_name = (char *)NAME_SYMV_U;
    }
    if ( 0 == strcmp(argv0,TEST_SYMV_L) ) {
      func      = SYMV_L;
      func_name = (char *)NAME_SYMV_L;
    }
#else
    if ( 0 == strcmp(argv0,TEST_HEMV_U) ) {
      func      = HEMV_U;
      func_name = (char *)NAME_HEMV_U;
    }
    if ( 0 == strcmp(argv0,TEST_HEMV_L) ) {
      func      = HEMV_L;
      func_name = (char *)NAME_HEMV_L;
    }
#endif
  }

  if ( argc > 1 ) {
    fp=fopen(argv[1], "r");
  } else {
    fp=fopen("IN", "r");
  }
  if ( fp == NULL ) {
    fprintf(stderr,"Cannot open parameter file.\n");
    exit(1);
  }

  fscanf(fp,"%d", &N);
  if ( N == 0 ) {
    mode = 'U';
  } else {
    mode= 'U';
  }

#if 0
  if ( argc > 2 ) {
    sscanf(argv[2], "%d", &blk);
    printf("% BLK specified=%d\n", blk);
  } else {
    blk = -1; //64;
  }
#endif

  print_head( func_name );

  int N_fault=0;
  int N_max = get_device_WorkSize( id, sizeof(TYPE) );

  while (1) {

    fscanf(fp,"%d", &N);
    if ( N <= 1 || N > N_max ) break;


    int     M = N;
#if 1
    if ( argc <= 3 )
      //if ( 0 )
      {
        if ( sizeof(TYPE) == 4 ) {
          if ( M < 12600 ) M = 12600;
        }
        if ( sizeof(TYPE) == 8 ) {
          if ( M < 9000 ) M = 9000;
        }
        if ( sizeof(TYPE) == 16 ) {
          if ( M < 6300 ) M = 6300;
        }
        if ( sizeof(TYPE) == 32 ) {
          if ( M < 4500 ) M = 4500;
        }
      }
#endif


    int     LDM = ((M-1)/(256/sizeof(TYPE))+1);
    LDM *= (256/sizeof(TYPE));
    LDA = ((N-1)/(256/sizeof(TYPE))+1);
    LDA *= (256/sizeof(TYPE));


    //	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
    cudaMallocHost((void **)&A, sizeof(TYPE) *(M*LDM));
    cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));

    //	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
    cublasAlloc(M*LDM, sizeof(TYPE), (void**)&dA);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);

    if(dA==NULL||dB==NULL||dC==NULL||dD==NULL) {
      fprintf(stderr, "# Fail to allocate memory %x %x %x %x\n",
              dA, dB, dC, dD);
      if(dA!=NULL) cublasFree(dA);
      if(dB!=NULL) cublasFree(dB);
      if(dC!=NULL) cublasFree(dC);
      if(dD!=NULL) cublasFree(dD);
      N_fault++;
      goto FFF;
    } else {
      N_fault=0;
    }




    printf("N= %d ", N);
    fflush(stdout);

    set_matrix( func, N, A, LDA, B, C, D );
    //	cublasSetMatrix(N, N, sizeof(TYPE), A, LDA, dA, LDA);
    cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
    cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
    cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
    cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
    cudaThreadSynchronize();



    double	tt[ITER];
    for(i=0;i<ITER;i++){

      cudaThreadSynchronize();
      double	tt1 = get_wtime();

      int	incb = 1;
      int	incc = 1;
      char	*opt = (char *)"U";

      switch (func) {
      case GEMV_N:
        magmablasgemv
          ( MagmaNoTrans, N, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
      case GEMV_T:
        magmablasgemv
          ( MagmaTrans, N, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
#if TYPE_DOUBLE || TYPE_FLOAT
      case SYMV_U:
        magmablassymv
          ( MagmaUpper, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
      case SYMV_L:
        magmablassymv
          ( MagmaLower, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
#else
      case HEMV_U:
        magmablashemv
          ( MagmaUpper, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
      case HEMV_L:
        magmablashemv
          ( MagmaLower, N, alpha, dA, LDA, dB, 1, beta, dD, 1);
        break;
#endif
      }

      cudaThreadSynchronize();
      double	tt2 = get_wtime();
      tt[i]=tt2-tt1;

    }
  EEE: fflush(stdout);


    { // sorting
      for(i=0;i<ITER;i++) 
        for(int j=i+1;j<ITER;j++)
          if(tt[i]>tt[j]){ double s=tt[i]; tt[i]=tt[j]; tt[j]=s; }

      // stats
      double	tmax=tt[ITER-1];
      double	t = tt[(ITER-1)/2];
      double	tvar=0.0;
      for(i=0;i<ITER;i++) tvar+=(tt[i]-t)*(tt[i]-t);
      tvar-=(tmax-t)*(tmax-t);
      tvar=sqrt(tvar)/(ITER-1);

      fflush(stdout);
      printf("%g [s] ", t);
#if TYPE_DOUBLE || TYPE_FLOAT
      printf("%g GFLOPS ", ((2.0*N)*N/t*1e-9));
#else
      printf("%g GFLOPS ", (4*(2.0*N)*N/t*1e-9));
#endif
      printf("%g ", tvar);
      if ( tvar/t > 2E-2 ) {
        printf("**%g**", tvar/t );
      }
      printf("%d ", LDA);
      fflush(stdout);
    }


  GGG: fflush(stdout);
    cublasFree(dA); //printf(":dA");
    cublasFree(dB); //printf(":dB");
    cublasFree(dC); //printf(":dC");
    cublasFree(dD); //printf(":dC");


  FFF: fflush(stdout);
    cudaFreeHost(A); //printf(":A");
    cudaFreeHost(B); //printf(":B");
    cudaFreeHost(C); //printf(":C");
    cudaFreeHost(D); //printf(":D");


  HHH: fflush(stdout);
    printf("\n");

    if ( N_fault > 4 ) break;
    if ( feof(fp) ) break;
  }
  fflush(stdout);


  cublasDestroy( handle );
  fclose(fp);
}

