#if TYPE_DOUBLE_COMPLEX || TYPE_FLOAT_COMPLEX
#include "aspen_complex.h"
#endif

#if TYPE_DOUBLE
#   define TYPE			double
#   define CONST(v)		((double)(v))
#   define Conj(v)		((double)(v))
#   define __isDOUBLE__			1
#   define __isFLOAT__			0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.7764e-15)
#   define _FUNCS		D_FUNCS
#endif
#if TYPE_FLOAT
#   define TYPE			float
#   define CONST(v)		((float)(v))
#   define Conj(v)		((float)(v))
#   define __isDOUBLE__			0
#   define __isFLOAT__			1
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.1921e-07)
#   define _FUNCS		S_FUNCS
#endif
#if TYPE_DOUBLE_COMPLEX
#   define TYPE			cuDoubleComplex
#   define CONST(v)		make_cuDoubleComplex(v,0.)
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDOUBLE_COMPLEX__		1
#   define __isFLOAT_COMPLEX__		0
#   define EPS			(1.7764e-15)
#   define _FUNCS		Z_FUNCS
#endif
#if TYPE_FLOAT_COMPLEX
#   define TYPE			cuFloatComplex
#   define CONST(v)		make_cuFloatComplex(v,0.)
#   define __isDOUBLE__			0
#   define __isFLOAT__			0
#   define __isDOUBLE_COMPLEX__		0
#   define __isFLOAT_COMPLEX__		1
#   define EPS			(1.1921e-07)
#   define _FUNCS		C_FUNCS
#endif

// macro for the legacy cublas
#define	cublasAlloc( a, b, c )	cudaMalloc ( (c), size_t((a)*(b)) )
#define	cublasFree( a )		cudaFree ( a )



#ifdef __cplusplus
extern "C" {
#endif

  double get_wtime(void);
  int do_usleep(int usec);
  int get_gpu_id(void);
  void print_head(char *func_name);

  void
  ASPEN_init( int );

  void
  ASPEN_shutdown( void );

  int
  ASPEN_get_driver_Version( void );

  int
  ASPEN_get_runtime_Version( void );

  size_t
  get_device_WorkSize( int id, int size );

  void
  ASPEN_SGEMV_t(int blk, float alpha, float beta, float *dA, float *dB,
                float *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_SGEMV_n(int blk, float alpha, float beta, float *dA, float *dB,
                float *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_SSYMVu(int blk, int N, float *dA, int LDA, float *dB,
               float *dC,
               float alpha, float beta);

  void
  ASPEN_SSYMVl(int blk, int N, float *dA, int LDA, float *dB,
               float *dC,
               float alpha, float beta);

  void
  ASPEN_DGEMV_t(int blk, double alpha, double beta, double *dA, double *dB,
                double *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_DGEMV_n(int blk, double alpha, double beta, double *dA, double *dB,
                double *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_DSYMVu(int blk, int N, double *dA, int LDA, double *dB,
               double *dC,
               double alpha, double beta);
  void
  ASPEN_DSYMVl(int blk, int N, double *dA, int LDA, double *dB,
               double *dC,
               double alpha, double beta);

  void
  ASPEN_CGEMV_t(int blk, cuFloatComplex alpha, cuFloatComplex beta, cuFloatComplex *dA, cuFloatComplex *dB,
                cuFloatComplex *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_CGEMV_n(int blk, cuFloatComplex alpha, cuFloatComplex beta, cuFloatComplex *dA, cuFloatComplex *dB,
                cuFloatComplex *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_CHEMVu(int blk, int N, cuFloatComplex *dA, int LDA, cuFloatComplex *dB,
               cuFloatComplex *dC,
               cuFloatComplex alpha, cuFloatComplex beta);
  void
  ASPEN_CHEMVl(int blk, int N, cuFloatComplex *dA, int LDA, cuFloatComplex *dB,
               cuFloatComplex *dC,
               cuFloatComplex alpha, cuFloatComplex beta);


  void
  ASPEN_ZGEMV_t(int blk, cuDoubleComplex alpha, cuDoubleComplex beta, cuDoubleComplex *dA, cuDoubleComplex *dB,
                cuDoubleComplex *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_ZGEMV_n(int blk, cuDoubleComplex alpha, cuDoubleComplex beta, cuDoubleComplex *dA, cuDoubleComplex *dB,
                cuDoubleComplex *dC,
                int LDA,
                int N1, int N2);
  void
  ASPEN_ZHEMVu(int blk, int N, cuDoubleComplex *dA, int LDA, cuDoubleComplex *dB,
               cuDoubleComplex *dC,
               cuDoubleComplex alpha, cuDoubleComplex beta);
  void
  ASPEN_ZHEMVl(int blk, int N, cuDoubleComplex *dA, int LDA, cuDoubleComplex *dB,
               cuDoubleComplex *dC,
               cuDoubleComplex alpha, cuDoubleComplex beta);

#ifdef __cplusplus
}
#endif

enum S_FUNCS { SGEMV_N, SGEMV_T, SSYMV_U, SSYMV_L };
enum D_FUNCS { DGEMV_N, DGEMV_T, DSYMV_U, DSYMV_L };
enum C_FUNCS { CGEMV_N, CGEMV_T, CHEMV_U, CHEMV_L };
enum Z_FUNCS { ZGEMV_N, ZGEMV_T, ZHEMV_U, ZHEMV_L };

