#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>
#include <complex.h>
#include <omp.h>
extern "C" {
#include <cblas.h>
}
#include <cuda_runtime.h>
#include <cublas_v2.h>

#define	ITER	1

#define USE_CUDA_EVENT_TIMER	0
#define CHECK_ANSWER		1
#include "test-common.h"
#include "test.h"

int
set_matrix_vectors( const int N, TYPE *A, const int lda, TYPE *X, TYPE *Y,
                    const enum _FUNCS func );


int
set_matrix_vectors( const int N, TYPE *A, const int lda, TYPE *X, TYPE *Y,
                    const enum _FUNCS func )
{

  srand(0);

#if TYPE_DOUBLE || TYPE_FLOAT
  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      A[j+i*lda]=(TYPE)(0);
    }
  }
#else
  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      A[j+i*lda].x=A[j+i*lda].y=0;
    }
  }
#endif

#if TYPE_DOUBLE || TYPE_FLOAT
  for(int i=0;i<N;i++) X[i]=(TYPE)1;
  for(int i=0;i<N;i++) Y[i]=(TYPE)0;
#else
  for(int i=0;i<N;i++) { X[i].x=1; X[i].y=0; }
  for(int i=0;i<N;i++) { Y[i].x=0; Y[i].y=0; }
#endif

#if TYPE_DOUBLE || TYPE_FLOAT
  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
      TYPE  t  = (TYPE)rand();
      A[j+lda*i] = t / RAND_MAX;
    }
  }

  if ( func == SYMV_U ) {
    for(int i=0;i<N;i++) {
      for(int j=0;j<i;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j] = (TYPE)0;
        A[j+lda*i] = (t1+t2)/2;
      }
    }
  }
  if ( func == SYMV_L ) {
    for(int i=0;i<N;i++) {
      for(int j=i+1;j<N;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j] = (TYPE)0;
        A[j+lda*i] = (t1+t2)/2;
      }
    }
  }
#else
  for(int i=0;i<N;i++) {
    for(int j=0;j<N;j++) {
#if TYPE_DOUBLE_COMPLEX
      double  t  = (double)rand();
      double  s  = (double)rand();
#endif
#if TYPE_FLOAT_COMPLEX
      float  t  = (float)rand();
      float  s  = (float)rand();
#endif
      A[j+lda*i].x = t / RAND_MAX;
      A[j+lda*i].y = s / RAND_MAX;
    }
  }

  if ( func == HEMV_U ) {
    for(int i=0;i<N;i++) {
      A[i+lda*i].y = 0;
      for(int j=0;j<i;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j].x = A[i+lda*j].y = 0;
        A[j+lda*i].x = (t1.x+t2.x)/2;
        A[j+lda*i].y = (t1.y+t2.y)/2;
      }
    }
  }
  if ( func == HEMV_L ) {
    for(int i=0;i<N;i++) {
      A[i+lda*i].y = 0;
      for(int j=i+1;j<N;j++) {
        TYPE  t1  = A[j+lda*i];
        TYPE  t2  = A[i+lda*j];
        A[i+lda*j].x = A[i+lda*j].y = 0;
        A[j+lda*i].x = (t1.x+t2.x)/2;
        A[j+lda*i].y = (t1.y+t2.y)/2;
      }
    }
  }
#endif
}


int
main(int argc, char *argv[])
{
  TYPE	alpha, beta;
#if TYPE_DOUBLE || TYPE_FLOAT
  alpha = (TYPE)3;
  beta  = (TYPE)0;//4.0;
#else
  alpha.x = 3;
  beta.x  = 0;//4.0;
  alpha.y = 0;
  beta.y  = 0;//4.0;
#endif
  TYPE	*A, *B, *C, *D;
  TYPE	*dA, *dB, *dC, *dD;
  int	N, M, K;
  int	LDA, LDB, LDC;
  int	i;
  int	blk;
  FILE	*fp;
#if CHECK_ANSWER
  int	verify = 0;
#endif
  int	id = get_gpu_id();

  enum _FUNCS func = GEMV_N;
  char	*func_name = (char *)"";

  cudaError_t st;


  if ( cudaSuccess != (st = cudaSetDevice(id)) ) {
    exit(1);
  }

  cublasStatus_t stat;
  cublasHandle_t handle;
  stat = cublasCreate( &handle );


  ASPEN_init( id );

  {
    char *argv0 = argv[0];
    if ( strrchr(argv0, '/') != argv0 ) {
      argv0 = strrchr(argv0,'/')+1;
    }
#if TYPE_DOUBLE || TYPE_FLOAT
    if ( 0 == strcmp(argv0,TEST_GEMV_N) ) {
      func      = GEMV_N;
      func_name = (char *)NAME_GEMV_N;
    }
    if ( 0 == strcmp(argv0,TEST_GEMV_T) ) {
      func      = GEMV_T;
      func_name = (char *)NAME_GEMV_T;
    }
#endif
#if TYPE_DOUBLE || TYPE_FLOAT
    if ( 0 == strcmp(argv0,TEST_SYMV_U) ) {
      func      = SYMV_U;
      func_name = (char *)NAME_SYMV_U;
    }
    if ( 0 == strcmp(argv0,TEST_SYMV_L) ) {
      func      = SYMV_L;
      func_name = (char *)NAME_SYMV_L;
    }
#else
    if ( 0 == strcmp(argv0,TEST_HEMV_U) ) {
      func      = HEMV_U;
      func_name = (char *)NAME_HEMV_U;
    }
    if ( 0 == strcmp(argv0,TEST_HEMV_L) ) {
      func      = HEMV_L;
      func_name = (char *)NAME_HEMV_L;
    }
#endif
  }

  if ( argc > 1 ) {
    fp=fopen(argv[1], "r");
  } else {
    fp=fopen("IN", "r");
  }
  if ( fp == NULL ) {
    fprintf(stderr,"Cannot open parameter file.\n");
    exit(1);
  }

  fscanf(fp,"%d", &N);

  if ( argc > 2 ) {
    sscanf(argv[2], "%d", &blk);
    printf("% BLK specified=%d\n", blk);
  } else {
    blk = -1; //64;
  }

  {
    verify = 1;
  }

  print_head( func_name );

#if CHECK_ANSWER
  printf("Run in the %s mode\n",
         verify ? "verify" : "non-verify(fast)" );
  fflush(stdout);
#endif
  printf("*****************************************/-->\n");
  fflush(stdout);


  int N_fault=0;
  int N_max = get_device_WorkSize( id, sizeof(TYPE) );

  for(int Itr = 0; fscanf(fp,"%d", &N) > 0; Itr++ ) {

    if ( N <= 1 || N > N_max ) break;


    int	M = N;

    int	LDM = ((M-1)/(256/sizeof(TYPE))+1);
    LDM *= (256/sizeof(TYPE));
    LDA = ((N-1)/(256/sizeof(TYPE))+1);
    LDA *= (256/sizeof(TYPE));


    //	cudaMallocHost((void **)&A, sizeof(TYPE) *(N*LDA));
    cudaMallocHost((void **)&A, sizeof(TYPE) *(M*LDM));
    cudaMallocHost((void **)&B, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&C, sizeof(TYPE) *(N*1));
    cudaMallocHost((void **)&D, sizeof(TYPE) *(N*1));

    //	cublasAlloc(N*LDA, sizeof(TYPE), (void**)&dA);
    cublasAlloc(M*LDM, sizeof(TYPE), (void**)&dA);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dB);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dC);
    cublasAlloc(N*1, sizeof(TYPE), (void**)&dD);

    if(dA==NULL||dB==NULL||dC==NULL||dD==NULL) {
      fprintf(stderr, "# Fail to allocate Dev_array %dx%d, %d*4\n",
              N, LDA, N);
      fprintf(stderr, "# Fail to allocate memory %x %x %x %x\n",
              dA, dB, dC, dD);
      if(dA!=NULL) cublasFree(dA);
      if(dB!=NULL) cublasFree(dB);
      if(dC!=NULL) cublasFree(dC);
      if(dD!=NULL) cublasFree(dD);
      N_fault++;
      goto FFF;
    } else {
      N_fault=0;
    }

    printf("N= %d ", N);
    fflush(stdout);
    fflush(stdout);
    fflush(stdout);


    set_matrix_vectors( N, A, LDA, B, C, func );

    for(int i=0;i<N;i++) D[i] = C[i];


    if ( argc > 3 ) { ; } else {
      cublasSetMatrix(LDM, M, sizeof(TYPE), A, LDM, dA, LDM);
      cublasSetVector(N, sizeof(TYPE), B, 1, dB, 1);
      cublasSetVector(N, sizeof(TYPE), C, 1, dC, 1);
      cublasSetVector(N, sizeof(TYPE), D, 1, dD, 1);
    }

    fflush(stdout);
    cudaThreadSynchronize();

    for(int i=0;i<ITER;i++){

      int	incb = 1;
      int	incc = 1;
      char	*opt = (char *)"U";

      switch (func) {
#if TYPE_DOUBLE || TYPE_FLOAT
      case GEMV_N:
        ASPEN_GEMV_n (blk, alpha, beta, dA, dB, dC, LDA, N, N);
        break;
      case GEMV_T:
        ASPEN_GEMV_t (blk, alpha, beta, dA, dB, dC, LDA, N, N);
        break;
#endif
#if TYPE_DOUBLE || TYPE_FLOAT
      case SYMV_U:
        ASPEN_SYMVu (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
      case SYMV_L:
        ASPEN_SYMVl (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
#else
      case HEMV_U:
        ASPEN_HEMVu (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
      case HEMV_L:
        ASPEN_HEMVl (blk, N, dA, LDA, dB, dC, alpha, beta);
        break;
#endif
      }

#if 1
      cudaError_t cu_err = cudaGetLastError();
      if ( cu_err != cudaSuccess ) {
        printf("CUDA internal ERROR detected\n"); goto HHH;
      }
#endif

      if ( argc > 3 ) goto FIN;

#if CHECK_ANSWER
      if(verify){
        if(i==0&&N<50000){
          cudaThreadSynchronize();
          do_usleep(100);
          for(int j=0;j<N;j++) {
#if TYPE_DOUBLE || TYPE_FLOAT
            C[j] = 0;
#else
            C[j].x = C[j].y = 0;
#endif
          }
          cublasGetVector(N, sizeof(TYPE), dC, 1, C, 1);
          cudaThreadSynchronize();
          do_usleep(100);
          switch (func) {
#if TYPE_DOUBLE || TYPE_FLOAT
          case GEMV_N:
            cblas_gemv
              ( CblasColMajor, CblasNoTrans, N, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          case GEMV_T:
            cblas_gemv
              ( CblasColMajor, CblasTrans, N, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
#endif
#if TYPE_DOUBLE || TYPE_FLOAT
          case SYMV_U:
            cblas_symv
              ( CblasColMajor, CblasUpper, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
          case SYMV_L:
            cblas_symv
              ( CblasColMajor, CblasLower, N, alpha, A, LDA, B, 1, beta, D, 1);
            break;
#else
          case HEMV_U:
            cblas_hemv
              ( CblasColMajor, CblasUpper, N, (void *)&alpha, (void *)A, LDA, (void *)B, 1, (void *)&beta, (void *)D, 1);
            break;
          case HEMV_L:
            cblas_hemv
              ( CblasColMajor, CblasLower, N, (void *)&alpha, (void *)A, LDA, (void *)B, 1, (void *)&beta, (void *)D, 1);
            break;
#endif
          }
          do_usleep(100);


#if TYPE_DOUBLE || TYPE_FLOAT
#if __isDOUBLE__
#define	FABS	fabs
#else
#define	FABS	fabsf
#endif
          TYPE t1,t2,t3;
          TYPE tol = 32*EPS;

          t1 = t2 = t3 = 0;
          for(int j=0;j<N;j++){
            t1 += FABS(C[j]-D[j]);
            t2 += FABS(C[j]);
            t3 += FABS(D[j]);
          }
#else
#if TYPE_DOUBLE_COMPLEX
          double t1, t2, t3;
          const double tol = 32*EPS;
          double complex *P = (double complex *)C;
          double complex *Q = (double complex *)D;
#define	FABS	cabs
#endif
#if TYPE_FLOAT_COMPLEX
          float t1, t2, t3;
          const float tol = 32*EPS;
          float complex *P = (float complex *)C;
          float complex *Q = (float complex *)D;
#define	FABS	cabsf
#endif
          t1 = t2 = t3 = 0;
          for(int j=0;j<N;j++){
            t1 += FABS(P[j]-Q[j]);
            t2 += FABS(P[j]);
            t3 += FABS(Q[j]);
          }
#endif

#if 0
#if TYPE_DOUBLE_COMPLEX
          printf( " ASPEN = %g,%g %g,%g %g,%g\n",
                  creal(P[0]), cimag(P[0]),
                  creal(P[1]), cimag(P[1]),
                  creal(P[2]), cimag(P[2]) );
          printf( " CBLAS = %g,%g %g,%g %g,%g\n",
                  creal(Q[0]), cimag(Q[0]),
                  creal(Q[1]), cimag(Q[1]),
                  creal(Q[2]), cimag(Q[2]) );
#endif
#if TYPE_FLOAT_COMPLEX
          printf( "\n");
          printf( " A[1] = %g,%g %g,%g %g,%g\n",
                  A[0+0*LDA].x, A[0+0*LDA].y,
                  A[1+0*LDA].x, A[1+0*LDA].y,
                  A[2+0*LDA].x, A[2+0*LDA].y );
          printf( " A[2] = %g,%g %g,%g %g,%g\n",
                  A[0+1*LDA].x, A[0+1*LDA].y,
                  A[1+1*LDA].x, A[1+1*LDA].y,
                  A[2+1*LDA].x, A[2+1*LDA].y );
          printf( " A[3] = %g,%g %g,%g %g,%g\n",
                  A[0+2*LDA].x, A[0+2*LDA].y,
                  A[1+2*LDA].x, A[1+2*LDA].y,
                  A[2+2*LDA].x, A[2+2*LDA].y );
          printf( " B = %g,%g %g,%g %g,%g\n",
                  B[0].x, B[0].y,
                  B[1].x, B[1].y,
                  B[2].x, B[2].y );

          printf( " ASPEN = %g,%g %g,%g %g,%g\n",
                  crealf(P[0]), cimagf(P[0]),
                  crealf(P[1]), cimagf(P[1]),
                  crealf(P[2]), cimagf(P[2]) );
          printf( " CBLAS = %g,%g %g,%g %g,%g\n",
                  crealf(Q[0]), cimagf(Q[0]),
                  crealf(Q[1]), cimagf(Q[1]),
                  crealf(Q[2]), cimagf(Q[2]) );
#endif
#endif

          printf( " | ASPEN - CBLAS | = %g\n",         t1     );
          printf( " | ASPEN | = %g, | CBLAS | = %g\n", t2, t3 );
          printf( " TOL = %g ", tol*t3 );

          if ( t1 >= tol * t2 ) {
            printf( " \n" );
            printf( " ********* \n" );
            printf( " **ERROR** \n" );
            printf( " ********* \n" );
          } else {
            printf( " ++PASSED++\n" );
          }

          cudaThreadSynchronize();
          do_usleep(100);
        }
      }
#endif

    }

  EEE:;

    cudaThreadSynchronize();

    fflush(stdout);
  GGG:
    fflush(stdout);

    cublasFree(dA); //printf(":dA");
    cublasFree(dB); //printf(":dB");
    cublasFree(dC); //printf(":dC");
    cublasFree(dD); //printf(":dC");

  FFF:
    cudaFreeHost(A); //printf(":A");
    cudaFreeHost(B); //printf(":B");
    cudaFreeHost(C); //printf(":C");
    cudaFreeHost(D); //printf(":D");

    printf("\n");
    fflush(stdout);
    fflush(stdout);

    if ( N_fault > 4 ) break;
    if ( feof(fp) ) break;
  }


  cublasDestroy( handle );
  ASPEN_shutdown();
 HHH:
  cudaDeviceReset();
  fclose(fp);
 FIN:
  ;

}

