#include "../tuning/CURRENT_GPU"
#include "aspen_devel.h"
#include "aspen_complex.h"
#include "aspen_half.h"


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__global__ void 
API(AXPY_kernel) ( TYPE * y, const TYPE alpha, TYPE * x, const long n, const long incx, const long incy )
{
  long pos = threadIdx.x + MULTIPLE * BLOCK_SIZE * blockIdx.x;
  x += pos*incx; y += pos*incy;

  if ( pos+BLOCK_SIZE*MULTIPLE <= n ) {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      y[0] += alpha * x[0];
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; y += BLOCK_SIZE*incy; }
    }
  } else {
#pragma unroll
    for ( int i=0; i<MULTIPLE; ++i ) {
      if ( pos+BLOCK_SIZE*i < n ) y[0] += alpha * x[0];
      if ( MULTIPLE > 1 ) { x += BLOCK_SIZE*incx; y += BLOCK_SIZE*incy; }
    }
  }
}


template < class TYPE, int BLOCK_SIZE, int MULTIPLE >
__host__ void
API(AXPY_host) ( TYPE * y, const TYPE alpha, TYPE * x, const int n, const int incx, const int incy )
{
  if ( n <= 0 ) return;

  int t1 = BLOCK_SIZE;
  int tt = MULTIPLE*t1;
  int n_blk = n / tt;
  int n_res = ((n_blk*tt<n)?1:0);
  int g1 = n_blk + n_res;

  int dev = API(get_device_ID)();
  int p_dev; cudaGetDevice( &p_dev );
  if ( dev != p_dev ) cudaSetDevice( dev );

  cudaStream_t stream = API(get_Stream)();

  API(AXPY_kernel) < TYPE , BLOCK_SIZE, MULTIPLE >
    <<< g1, t1, 32, stream
    >>> ( y, alpha, x, (long) n, (long) incx, (long) incy );

  if ( dev != p_dev ) cudaSetDevice( p_dev );
}


#define BLOCK_SIZE	(16*12)
#define	MULTIPLE	(2)

#if ASPEN_HALF_ENABLED
extern "C" void
API(haxpy) ( const int n, const half alpha, half * x, const int incx, half *y, const int incy )
{
  API(AXPY_host) < half, BLOCK_SIZE, MULTIPLE > ( y, alpha, x, n, incx, incy );
}
#endif

extern "C" void
API(saxpy) ( const int n, const float alpha, float * x, const int incx, float *y, const int incy )
{
  API(AXPY_host) < float, BLOCK_SIZE, MULTIPLE > ( y, alpha, x, n, incx, incy );
}

extern "C" void
API(daxpy) ( const int n, const double alpha, double * x, const int incx, double * y, const int incy )
{
  API(AXPY_host) < double, BLOCK_SIZE, MULTIPLE > ( y, alpha, x, n, incx, incy );
}

extern "C" void
API(caxpy) ( const int n, cuFloatComplex alpha, cuFloatComplex * x, const int incx, cuFloatComplex * y, const int incy )
{
  API(AXPY_host) < cuFloatComplex, BLOCK_SIZE, MULTIPLE > ( y, alpha, x, n, incx, incy );
}

extern "C" void
API(zaxpy) ( const int n, cuDoubleComplex alpha, cuDoubleComplex * x, const int incx, cuDoubleComplex * y, const int incy )
{
  API(AXPY_host) < cuDoubleComplex, BLOCK_SIZE, MULTIPLE > ( y, alpha, x, n, incx, incy );
}

