#!/bin/sh

echo_Message()
{
	echo '<--/****************************************'
	echo 'Automatic Performance Tuning for CHEMV-L'
	date
	echo -n 'Host on '; echo `hostname`
	echo 'Device is '$DEVICE
	if [ x${1} = xbegin ]; then
		echo 'Start.'
	fi
	if [ x${1} = xend ]; then
		echo 'Successfully completed.'
	fi
	if [ x${1} = xterminate ]; then
		echo 'Terminated unfortunately.'
	fi
	echo '****************************************/-->'
}


export LANG=C
LOG=log-chemvl-AT1X.`date +%Y.%m.%d.%T`
WORK_DIR=chemvl-current

if [ x${CUDA_PATH} != x ]; then
        export LD_LIBRARY_PATH=$CUDA_PATH/lib64:$LD_LIBRARY_PATH
fi

cd ../src; make get_dev_info; cd ../tuning
../src/get_dev_info $ASPEN_GPU_ID > DEV_INFO

DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
MP=`awk '/MP=/{ print $2 }' DEV_INFO`
CG=`awk '/CG=/{ print $2 }' DEV_INFO`
MAXDIM=`awk '/MAXDIM=/{ print $2 }' DEV_INFO`

cd ../src; make get_mult get_mult32 get_mult48; cd ../tuning

echo_Message begin

/bin/sh ./touch_header.sh

mkdir $LOG
if [ -f $WORK_DIR ]; then
	\rm -rf $WORK_DIR
fi
if [ -h $WORK_DIR ]; then
	\rm $WORK_DIR
fi
if [ -d $WORK_DIR ]; then
	\rm -rf $WORK_DIR
fi
ln -s $LOG $WORK_DIR

echo 'Tuning processes are recorded on '$LOG'.'

cd $WORK_DIR
cd ../../src; make get_dev_info; cd ../tuning/$WORK_DIR
ln -s ../../src/get_dev_info .

tar -C ../.. -cf - src template | tar -xf -

cat hemv_lower_atomic_template.incl

/bin/sh ../d.sh > OK-pat-symv

touch done_pattern
\rm done_pattern

/bin/sh ../chemvl_a.sh | tee $LOG
/bin/sh ../chemvl_b.sh $LOG
/bin/sh ../chemvl_b2.sh $LOG-X
/bin/sh ../chemvl_c.sh
/bin/sh ../d_filter.sh chemv-lower-auto3.h chemv-lower-auto2.h

echo '#if 0'             > chemv-lower-auto_.h
echo_Message            >> chemv-lower-auto_.h
cat ../DEV_INFO         >> chemv-lower-auto_.h
echo '<--'              >> chemv-lower-auto_.h
cat ../CURRENT_GPU      >> chemv-lower-auto_.h
echo '-->'              >> chemv-lower-auto_.h
echo '#endif'           >> chemv-lower-auto_.h
cat chemv-lower-auto.h  >> chemv-lower-auto_.h
mv chemv-lower-auto_.h chemv-lower-auto.h
cp chemv-lower-auto.h ..

echo '#if 0'             > chemv-lower-auto_.h
echo_Message            >> chemv-lower-auto_.h
cat ../DEV_INFO         >> chemv-lower-auto_.h
echo '<--'              >> chemv-lower-auto_.h
cat ../CURRENT_GPU      >> chemv-lower-auto_.h
echo '-->'              >> chemv-lower-auto_.h
echo '#endif'           >> chemv-lower-auto_.h
cat chemv-lower-auto2.h >> chemv-lower-auto_.h
mv chemv-lower-auto_.h chemv-lower-auto2.h
cp chemv-lower-auto2.h ..

cat chemv-lower-auto.h
cat chemv-lower-auto2.h

cd ../../src

\rm chemv_lower-*.cu_o
make -j4

cd ../bench

\rm test-z.o
make

./test-chemv-l IN-medium
./test2-chemv-l IN-medium

cd ../tuning

touch .done-chemvl

echo_Message end

exit 0

