#!/bin/bash

Init()
{

	../../src/get_dev_info $ASPEN_GPU_ID > DEV_INFO

	DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
	MP=`awk '/MP=/{ print $2 }' DEV_INFO`
	CG=`awk '/CG=/{ print $2 }' DEV_INFO`
	MAXDIM=`awk '/MAXDIM3=/{ print $2 }' DEV_INFO`
	CUDA=`awk '/CUDA=/{ print $2 }' DEV_INFO`

	\rm DEV_INFO

	if [ -e ../CURRENT_GPU ]; then
	CG_=`awk '{print $3*1}' ../CURRENT_GPU`
	if [ $CG -ne $CG_ ]; then
		\rm ../CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > ../CURRENT_GPU
	fi
	else
		\rm ../CURRENT_GPU
		echo '#define CURRENT_GPU '$CG > ../CURRENT_GPU
	fi

	if [ -f IN-exe-a ]; then
		\rm IN-exe-a
	fi
	cat << EOS_DO_A >> IN-exe-a
	1
	128
	256
	512
	1001
	2001
	2501
	3001
	3301
	4001
	7001
	10001
EOS_DO_A
	N=13000
	N_STEP=3000
	while [ $N -lt $MAXDIM ]
	do
		echo $N >> IN-exe-a
		D_STEP=`expr $N_STEP "*" 33`
		D_STEP=`expr $D_STEP "/" 100`
		N_STEP=`expr $N_STEP "+" $D_STEP`
		N=`expr $N "+" $N_STEP`
	done
		echo $MAXDIM >> IN-exe-a 
	echo 0 >> IN-exe-a

	if [ -f IN-exe-b ]; then
		\rm IN-exe-b
	fi
	touch IN-exe-b
	N=1
	while [ $N -lt 1001 ]
	do
		echo $N >> IN-exe-b
		N=`expr $N "+" 50`
	done
	N=1001
	NSTEP=128
	if [ $MAXDIM -gt 10000 ]; then
	while [ $N -lt 10000 ]
	do
		echo $N >> IN-exe-b
		N=`expr $N "+" $NSTEP`
		NSTEP=`expr $NSTEP "*" 101`
		NSTEP=`expr $NSTEP "/" 100`
	done
	fi
	if [ $MAXDIM -gt 20000 ]; then
	while [ $N -lt 20000 ]
	do
		echo $N >> IN-exe-b
		N=`expr $N "+" $NSTEP`
		NSTEP=`expr $NSTEP "*" 104`
		NSTEP=`expr $NSTEP "/" 100`
	done
	fi
	while [ $N -lt $MAXDIM ]
	do
		echo $N >> IN-exe-b
		N=`expr $N "+" $NSTEP`
		NSTEP=`expr $NSTEP "*" 107`
		NSTEP=`expr $NSTEP "/" 100`
	done
	echo $MAXDIM >> IN-exe-b 
	echo 0 >> IN-exe-b

	UX_PATTERN=""
	M=1; while [ $M -lt 193 ]
	do
		UX_PATTERN=$UX_PATTERN" "$M
		M=`expr $M "+" 1`
	done

	if [ $CG -lt 300 ]; then
		MM_PATTERN=" 8 7 6 5 4 3 2 "
		MM_MAX=8
	else
	if [ $CG -lt 500 ]; then
		MM_PATTERN=" 16 15 14 13 12 11 10 9 8 7 6 5 4 3 2 "
		MM_MAX=16
	else
		MM_PATTERN=" 32 31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 9 8 7 6 5 4 3 2 "
		MM_MAX=32
	fi
	fi

	Gmin=64
	Gmax=512

  if [ x$ASPEN_TUNING_LEVEL = xROUGH ]; then
	Gmin=128
	Gmax=384
  fi
  if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
	Gmin=32
	Gmax=768
  fi

}

Setup()
{
	if [ -f zhemv-upper-auto.h ]; then
		\rm zhemv-upper-auto.h
	fi
	awk ' \
		BEGIN{ \
			print "switch (BLK) {"; \
			for ( i=1; i <= 10*'$MM_MAX'; i++ ) { \
				print "#if __KERNEL"i; \
				print "case "i":"; \
				print "\tAPI(HEMVu_ATOMIC_host) < GPU_ARCH, scalar_t, BLOCK_SIZE, VX, UX, "int((i-1)/10)+1", "(i-1)%10" >"; \
				print "\t( n, a, lda, x, y, alpha, beta ); break;"; \
				print "#endif" \
			} \
			print "default:"; \
			print "\tfprintf( stderr, \"Not proper Parameter %d \", BLK );"; \
			print "\tbreak;"; \
			print "}"; \
			exit; \
		} { exit; } END { exit; }' > zhemv-upper-auto.h
	cp zhemv-upper-auto.h ../
}

Make0()
{
	if [ -f param-zhemvu.h ]; then
		\rm param-zhemvu.h
	fi
	touch param-zhemvu.h

cat << EOF_1 > param-zhemvu.h
#define	BLOCK_SIZE	$BLOCK_SIZE
#define	VX		$VX
#define	UX		$UX

// $MMM_PATTERN
EOF_1

	for MULTI in \
		$MMM_PATTERN
	do
	for MX in \
		0 1 2 3 4 5 6 7 8 9
	do
		ID=`expr $MULTI "-" 1`
		ID=`expr $ID "*" 10`
		ID=`expr $ID "+" $MX`
		ID=`expr $ID "+" 1`
		echo "#define	__KERNEL"$ID" 1" >> param-zhemvu.h
	done
	done
	if [ -f ../param-zhemvu.h ]; then
		\rm ../param-zhemvu.h
	fi
	cp param-zhemvu.h ../
}

Make1()
{
	if [ -f param-zhemvu.h ]; then
		\rm param-zhemvu.h
	fi
	touch param-zhemvu.h

cat << EOF_2 > param-zhemvu.h
#define	BLOCK_SIZE	$BLOCK_SIZE
#define	VX		$VX
#define	UX		$UX

// $MMM_PATTERN
EOF_2

	for MULTI in \
		$MMM_PATTERN
	do
	for MX in \
		$MX0
	do
		ID=`expr $MULTI "-" 1`
		ID=`expr $ID "*" 10`
		ID=`expr $ID "+" $MX`
		ID=`expr $ID "+" 1`
		echo "#define	__KERNEL"$ID" 1" >> param-zhemvu.h
	done
	done
	if [ -f ../param-zhemvu.h ]; then
		\rm ../param-zhemvu.h
	fi
	cp param-zhemvu.h ../
}

Make()
{
	cd ../../src
	\rm zhemv_upper-*_template.cu_o
	make -j4
	cd ../bench
	make
	cd ../tuning/zhemvu-current
}

Do()
{
cat << EOF_Do
#define	BLOCK_SIZE	$BLOCK_SIZE
#define	VX		$VX
#define	UX		$UX
#define	MULTIPLICITY	$MULTI
#define	_M_		$MX
EOF_Do

	MX1=`expr $MX + 1`
	MX2=`expr $MULTI - 1`
	MX2=`expr $MX2 "*" 10`
	MX3=`expr $MX1 + $MX2`

cat << EOF_INnv > IN-nv
1
1001
6001
9001
EOF_INnv
awk '/MAXDIM=/{ print $2; }' ../DEV_INFO >> IN-nv
echo '-1' >> IN-nv

  if [ x$ASPEN_CHECK_ERROR = x1 ]; then
	timeout -s KILL 120 ../../bench/test-zhemv-l IN-nv $MX3 | \
		awk ' \
			/cannot free/{ \
				system("touch ./.NG"); \
			} \
			/ERR/{ \
				system("touch ./.NG"); \
			} \
			/unavailable/{ \
				system("touch ./.NG"); \
			} \
			{ \
				print;fflush(); \
			} \
		' > /dev/null
  fi


	if [ ! -e ./.NG ]; then

		timeout -s KILL 720 ../../bench/test2-zhemv-u IN-exe-a $MX3 | tee trial-$MX3
		echo $BLOCK_SIZE" "$VX" "$UX" "$MULTI" "$MX >> done_pattern

		cat trial-$MX3 \
		| awk '/N=/{ print $2" "$5 }' \
		> data-$MX3

		cat MAXF-$BLOCK_SIZE-$VX-$MULTI data-$MX3 \
		| awk '\
			function fabs(x) { return (x>0?x:-x); } \
			BEGIN{ stable=1; } \
			{ N=$1; P=fabs($2); C=-pp[N]; \
				if ( C==0 ) { \
					pp[N] = -P; \
				} else if ( C < P ) { \
					print "*** "N" "C" < "P; \
					pp[N] = -P; stable = 0; \
				} \
			} \
			END { \
				for ( N in pp ) { \
					print N" "pp[N]; \
				} \
				if ( stable ) { \
					system( "touch NG2"); \
				} \
			}' \
		| sort -gk 1 \
		> MAXF2-$BLOCK_SIZE-$VX-$MULTI

		cat MAXF2-$BLOCK_SIZE-$VX-$MULTI \
			| grep "<" \
			| sort -gk 2

		cat MAXF2-$BLOCK_SIZE-$VX-$MULTI \
			| grep -v "<" \
			> MAXF-$BLOCK_SIZE-$VX-$MULTI


	fi
}



Init

Setup

echo "Sampling start"

	GLOBAL_MAX_UX_MX=0

if [ -L ../zhemvu_reguse ]; then
	\rm ../zhemvu_reguse
fi

if [ x$ASPEN_REGUSE_CALC != x1 ]; then
	if [ -f ../zhemvu_reguse-orig.tgz ]; then
		tar -zxf ../zhemvu_reguse-orig.tgz
		mv ./zhemvu_reguse-orig ..
	fi
	ln -s ./zhemvu_reguse-orig ../zhemvu_reguse
fi

if [ -e ../zhemvu_reguse/$CUDA/$CG/GLOBAL_MAX_UX_MX ]; then

	cp ../zhemvu_reguse/$CUDA/$CG/* .
	cat log-reguse-all
	GLOBAL_MAX_UX_MX=`cat GLOBAL_MAX_UX_MX`

else

    VX_PATTERN="1 2 3 4 5 6 7 8"
  if [ x$ASPEN_TUNING_LEVEL = xROUGH ]; then
    VX_PATTERN="2 3 4 5 6 7 8"
  fi

for VX in \
	$VX_PATTERN
do
for BLOCK_SIZE in \
	32 64 96 128 160 192 224 256
do
if [ `expr $BLOCK_SIZE "*" $VX` -ge $Gmin ]; then
if [ `expr $BLOCK_SIZE "*" $VX` -le $Gmax ]; then

  touch log-regs-$BLOCK_SIZE
  \rm log-regs-$BLOCK_SIZE

	MAX_UX_MX=0
	last_MMM=9999

  for UX in \
	$UX_PATTERN
  do

  THREADS=`expr $BLOCK_SIZE "*" $VX`
  if [ $UX -le $THREADS ]; then

  if [ `expr $UX "%" $VX` -eq 0 ]; then
  if [ `expr $UX "/" $VX` -ge 1 ]; then
  if [ `expr $UX "/" $VX` -le 32 ]; then


	MMM_PATTERN="1"
	Make0 >& /dev/null

	M=`expr $UX "/" 2`
	M=`expr $M "*" 2`
	M=`expr $M "+" 1`
	M=`expr $M "*" $UX`
	M2=`expr $BLOCK_SIZE "*" $VX`
#	if [ $M -lt $M2 ]; then
		M=$M2
#	fi
#	M1=`expr $M "%" 2`
#	M=`expr $M "+" $M1`
	M=`expr $M "+" $UX`
	M=`expr $M "+" 8`
	SHMEM_MIN=`expr $M "*" 16`

	/bin/sh ../get_reguse.sh $CG main_do \
		../../src/zhemv_upper-X_template \
		../../src/zhemv_upper-X_template \
		../../src/zhemv_upper-X_template \
		../../src/zhemv_upper-X_template \
		../../src/zhemv_upper-X_template \
		../../src/zhemv_upper-X_template \
		| \
        sort -k 5 > log-reg-all

	MMM=0
	OOO=0
	CCC=0
	SSS=0
	XXX=0

	SSS_1=""
	SSS_2=""

      for MX in \
	0 1 2 3 4 5 6 7 8 9
      do

	touch log-regs
	\rm log-regs

cat << EOF_IN1 > IN-1
1
6001
-1
EOF_IN1

	cat log-reg-all | awk '{if($5=='$MX')print $1}' > CX
	CX=`cat CX`

	cat log-reg-all | awk '{if($5=='$MX')print $7}' > SX
	SX=`cat SX`

	M=`expr $BLOCK_SIZE "*" $VX`
if [ $CG -lt 200 ]; then
	../../src/get_mult $M $CX $SHMEM_MIN $CG > multi_use
else
	../../src/get_mult48 $M $CX $SHMEM_MIN $CG > multi_use
fi

	MMM_PATTERN=`cat multi_use | \
		awk '{ gsub(/\([0-9]*\)/," "); print; }'`
	OOO_PATTERN=`cat multi_use | \
		awk '{ gsub(/[0-9]*\(/," "); gsub(/\)/,""); print; }'`
	M=`echo $MMM_PATTERN | awk '{ print $(NF); }'`
	O=`echo $OOO_PATTERN | awk '{ print $(NF); }'`
	if [ $M -gt $MMM ]; then
		if [ ! $MMM -eq 0 ]; then
			CCC=1
		fi
		MMM=$M
		OOO=$O
		XXX=$MX

		SSS_1=""
		SSS_2=""
	else
		if [ ! $M -eq $MMM ]; then
			CCC=1
		fi
	fi

	UX_MX=`expr $UX "*" $M`
if [ $UX_MX -gt $MAX_UX_MX ]; then
	if [ $SX -lt 2 ] && [ $BLOCK_SIZE -ge $UX ]; then
		MAX_UX_MX=$UX_MX
	fi
fi

if [ $M -eq $MMM ]; then

	if [ $SX -eq 1 ]; then
		SSS_1=$MX" "$SSS_1
	fi
	if [ $SX -eq 2 ]; then
		SSS_2=$MX" "$SSS_2
	fi

fi

	if [ $SSS -lt $SX ]; then
		SSS=$SX
	fi

	echo $BLOCK_SIZE' '$VX' '$UX' '$MX' Register= '$CX ' / '$SX' :: '$SHMEM_MIN' : '$MMM_PATTERN | \
		tee -a log-regs-$BLOCK_SIZE

      done # for MX

	if [ ! "x$SSS_1" = "x" ]; then
		XXX=`echo $SSS_1 | awk '{print $1}'`
	else
	if [ ! "x$SSS_2" = "x" ]; then
		XXX=`echo $SSS_2 | awk '{print $1}'`
	else
		XXX=0
	fi
	fi

	echo '## '$BLOCK_SIZE' '$VX' '$UX' '$MMM' / '$OOO'% ! '$CCC' > '$SSS '::' $XXX | \
		tee -a log-regs-$BLOCK_SIZE


	if [ $MMM -le 1 ]; then
	if [ $last_MMM -le 1 ]; then
		break
	fi
	fi
	if [ $SSS -eq 0 ]; then
		break
	fi
	last_MMM=$MMM

  fi
  fi
  fi

  fi
  done # for UX

M=`expr $MAX_UX_MX "*" $BLOCK_SIZE`
if [ $M -gt $GLOBAL_MAX_UX_MX ]; then
	GLOBAL_MAX_UX_MX=$M
fi

  cp log-regs-$BLOCK_SIZE log-regs-$BLOCK_SIZE-$VX
  echo $MAX_UX_MX > MAX_UX_MX-$BLOCK_SIZE-$VX

fi
fi
done # for BLOCK_SIZE
done # for VX

echo $GLOBAL_MAX_UX_MX > GLOBAL_MAX_UX_MX

fi

memhog 4gb


if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
	GLOBAL_MAX_UX_MX=10000000
else
# 64*64*4=16384
if [ $GLOBAL_MAX_UX_MX -gt 16384 ]; then
	GLOBAL_MAX_UX_MX=16384
fi
fi

    VX_PATTERN="1 2 3 4 5 6 7 8"
  if [ x$ASPEN_TUNING_LEVEL = xROUGH ]; then
    VX_PATTERN="2 3 4 5 6 7 8"
  fi

for VX in \
	$VX_PATTERN
do
for BLOCK_SIZE in \
	32 64 96 128 160 192 224 256
do
if [ `expr $BLOCK_SIZE "*" $VX` -ge $Gmin ]; then
if [ `expr $BLOCK_SIZE "*" $VX` -le $Gmax ]; then

  if [ -e MAX_UX_MX-$BLOCK_SIZE-$VX ]; then
    MAX_UX_MX=`cat MAX_UX_MX-$BLOCK_SIZE-$VX`
    echo 'MAX UX*MX='$MAX_UX_MX' GLOBAL:MAX_BLK*UX*MX='$GLOBAL_MAX_UX_MX
  else
    MAX_UX_MX=1000000
  fi

  UU_PATTERN_old=""

  DONE=0
  for MULTI in \
	$MM_PATTERN
  do

	if [ $DONE -gt 6 ]; then
		break
	fi

	MM16_PATTERN=" 32 21 16 12 10 9 8 7 6 5 4 3 2 1"
	MM48_PATTERN=" 32 27 24 21 19 17 16 14 13 12 11 10 9 8 7 6 5 4 3 2 1"

	MMxx_PATTERN=$MM48_PATTERN

	M_C=0
	for M in \
		$MMxx_PATTERN
	do
		if [ $M -eq $MULTI ]; then
			M_C=`expr $M_C "+" 1`
		fi
	done # for M

	MMM_PATTERN=$MULTI


	if [ $M_C -eq 0 ]; then
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 == '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -4 | awk '{printf("%d ",$0) }' > UUU
	else
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 >= '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -4 | awk '{printf("%d ",$0) }' > UUU
	fi

	NUU=`awk '{ if(NF>0)print $1; else print 0; }END{ if(NR==0)print 0; }' UUU`
	NLL=`awk '{ if(NF>8)print $8; else if(NF>0) print $NF; else print 1000; }END{ if(NR==0)print 1000; }' UUU`

	if [ $CG -le 200 ]; then
		NK=8
	else
		NK=4
	fi
	if [ $BLOCK_SIZE -lt $NUU ]; then
		NK=8
	fi

if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
		NK=256
fi

	if [ $M_C -eq 0 ]; then
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 == '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -$NK | awk '{printf("%d ",$0) }' > UUU
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 == '$MULTI' && $11 ) print $9 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > CCC
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk 'BEGIN{print 1}/##/{ if ( $5 == '$MULTI' && $11 ) print $11 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > SSS
	else
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 >= '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -$NK | awk '{printf("%d ",$0) }' > UUU
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $5 >= '$MULTI' && $11 ) print $9 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > CCC
	cat log-regs-$BLOCK_SIZE-$VX | \
		awk 'BEGIN{print 1}/##/{ if ( $5 >= '$MULTI' && $11 ) print $11 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > SSS
	fi


	UU_PATTERN=`cat UUU`
	CCC=`cat CCC`
	SSS=`cat SSS`

	if [ $BLOCK_SIZE -lt $NUU ]; then
	if [ $BLOCK_SIZE -lt $NLL ]; then

		NK=10
if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
		NK=256
fi
		M1=`expr $BLOCK_SIZE "/" $VX`
		M1=`expr $M1 "*" $VX`
		M2=`expr $M1 "-" $VX`

		echo $UU_PATTERN | \
			awk ' \
			$0 !~ /[\y ]'$M1'[\y ]/{ printf(" %s ",'$M1') } \
			$0 !~ /[\y ]'$M2'[\y ]/{ printf(" %s ",'$M2') } \
		' >> UUU

		UU_PATTERN=`cat UUU`

	fi
	fi

	if [ $SSS -gt 0 ]; then
	# parameters are the first CCC+1 UX from UU_PATTERN
	if [ $SSS -gt 1 ]; then
		#CCC=3
		CCC=`expr $NK "-" 1`
	else
	if [ $BLOCK_SIZE -lt $NUU ]; then
		CCC=4
	else
		#CCC=1
		CCC=2
	fi
	fi
	fi
if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
		CCC=256
fi

	M1=`echo $UU_PATTERN     | awk '{ gsub(/ /,":"); print; }'`

	if [ x${M1} = x ]; then
		echo 'UX[MULTI='$MULTI']:=<none>'
	else
		M=`expr $CCC "+" 1`
		echo 'UX[MULTI='$MULTI']:=<'$UU_PATTERN'> <= '$M
	fi

	#M2=`echo $UU_PATTERN_old | awk '{ gsub(/ /,":"); print; }'`
	M2=""

if [ x$ASPEN_TUNING_LEVEL != xFULL ] && [ x${M1} = x${M2} ]; then
	echo '** Skip'
else

	GTOL=33
	LTOL=50

if [ x$ASPEN_TUNING_LEVEL = xFULL ]; then
	GTOL=0
	LTOL=0
fi

if [ x$ASPEN_TUNING_LEVEL = xROUGH ]; then
	GTOL=50
	LTOL=75
fi



	M3=`echo $UU_PATTERN "0" | awk '{print $1}'`
	M4=`expr $MULTI "*" $M3`
	M5=`expr $MAX_UX_MX "*" $LTOL`
	M5=`expr $M5 "/" 100`
	M6=`expr $M4 "*" $BLOCK_SIZE`

if [ $M6 -gt $GLOBAL_MAX_UX_MX ]; then
	echo '** Skip, because BLK*MX*UX > Global:MAX(BLK*MX*UX)'
else

	M7=`expr $GLOBAL_MAX_UX_MX "*" $GTOL`
	M7=`expr $M7 "/" 100`

if [ $M6 -lt $M7 ]; then
	echo '** Skip, because BLK*MX*UX < 0.'$GTOL'*Global:MAX(BLK*MX*UX)'
else
if [ $M4 -lt $M5 ]; then
	echo '** Skip, because MX*UX < 0.'$LTOL'*Local:MAX(MX*UX)'
else

cat IN-exe-a | awk '{ print $1" -1e-10" }' > MAXF-$BLOCK_SIZE-$VX-$MULTI

    for UX in \
	$UU_PATTERN
    do

    THREADS=`expr $BLOCK_SIZE "*" $VX`
    if [ $UX -le $THREADS ]; then

    if [ `expr $UX "%" $VX` -eq 0 ]; then
    if [ x$ASPEN_TUNING_LEVEL = xFULL ] || [ `expr $UX "/" $VX` -ge 3 ]; then
    if [ `expr $UX "/" $VX` -le 32 ]; then

	M8=`expr $MULTI "*" $UX`
	M9=`expr $M8 "*" $BLOCK_SIZE`

if [ $M8 -lt $M5 ] || [ $M9 -lt $M7 ]; then

if [ $M9 -lt $M7 ]; then
	echo '** Skip, because BLK*MX*UX < 0.'$GTOL'*Global:MAX(BLK*MX*UX)'
else
if [ $M8 -lt $M5 ]; then
	echo '** Skip, because MX*UX < 0.'$LTOL'*Local:MAX(MX*UX)'
fi
fi

else

	cat log-regs-$BLOCK_SIZE-$VX | \
		awk '/##/{ if ( $3=='$VX' && $4=='$UX') print $13}' >  MX0
	MX0=`cat MX0`

	Make1 >& /dev/null
	Make  >& /dev/null

		if [ -e NG ]; then
			\rm NG
		fi

#      for MX in \
#	0 1 2 3 4 5 6 7 8 9
      for MX in \
	$MX0
      do

		if [ -e .NG ]; then
			\rm .NG
		fi
		if [ -e NG2 ]; then
			\rm NG2
		fi

		Do | awk ' \
			/ERROR/ { \
				print; \
				fflush(); \
				system("touch NG"); \
				exit; \
				next; \
			} \
			{ \
				print; \
				fflush(); \
			} \
			END { \
				fflush(); \
			} \
			'

		if [ -e .NG ]; then
			echo 'ERROR in kernel implementation!!'
			\rm .NG
			touch NG
		fi

		if [ -e NG ]; then
			break
		fi

      done # for MX

		if [ -e NG ]; then
			\rm NG
		else
			if [ $CCC -gt 0 ]; then
				CCC=`expr $CCC "-" 1`
			else
				DONE=`expr $DONE "+" 1`
				break
			fi
		fi

		if [ -e NG2 ]; then
			echo '***********'
			\rm NG2
			break
		fi

fi

    fi
    fi
    fi

    fi
    done # for UX

fi
fi
fi

fi

    UU_PATTERN_old=$UU_PATTERN

  done # for MULTI

fi
fi
done # for BLOCK_SIZE
done # for VX

echo "Complete phase a"
